'''

# #include "docstrings/docstring.dst.txt"

'''
from __future__ import absolute_import

__version__ = '1.5.00'
version = __version__
__author__ = 'Hans Petter Langtangen', 'Johannes H. Ring', 'Kristian Gregorius Hustad'
author = __author__

__acknowledgments__ = ''

from .doconce import doconce_format, DocOnceSyntaxError
